atis_mapping = {
    'depart_date.day_number':'<<depart date day number>>',
    'arrive_date.day_name':'<<arrive date day name>>',
    'airline_name':'<<airline name>>',
    'depart_date.year':'<<depart date year>>',
    'flight_mod':'<<flight mod>>',
    'return_date.day_name':'<<return date day name>>',
    'toloc.city_name':'<<toloc city name>>',
    'return_date.day_number':'<<return date day number>>',
    'time_relative':'<<time relative>>',
    'city_name':'<<city name>>',
    'state_code':'<<state code>>',
    'transport_type':'<<transport type>>',
    'class_type':'<<class type>>',
    'days_code':'<<days code>>',
    'toloc.country_name':'<<toloc country name>>',
    'arrive_date.today_relative':'<<arrive date today relative>>',
    'round_trip':'<<round trip>>',
    'toloc.state_name':'<<toloc state name>>',
    'aircraft_code':'<<aircraft code>>',
    'arrive_date.month_name':'<<arrive date month name>>',
    'depart_date.today_relative':'<<depart date today relative>>',
    'depart_time.start_time':'<<depart time start time>>',
    'compartment':'<<compartment>>',
    'day_number':'<<day number>>',
    'depart_date.date_relative':'<<depart date date relative>>',
    'arrive_date.day_number':'<<arrive date day number>>',
    'depart_time.time':'<<depart time time>>',
    'fare_amount':'<<fare amount>>',
    'depart_date.month_name':'<<depart date month name>>',
    'period_of_day':'<<period of day>>',
    'cost_relative':'<<cost relative>>',
    'fromloc.airport_name':'<<fromloc airport name>>',
    'fare_basis_code':'<<fare basis code>>',
    'arrive_time.start_time':'<<arrive time start time>>',
    'stoploc.airport_name':'<<stoploc airport name>>',
    'time':'<<time>>',
    'depart_time.time_relative':'<<depart time time relative>>',
    'return_time.period_of_day':'<<return time period of day>>',
    'depart_time.period_of_day':'<<depart time period of day>>',
    'economy':'<<economy>>',
    'mod':'<<mod>>',
    'stoploc.airport_code':'<<stoploc airport code>>',
    'stoploc.state_code':'<<stoploc state code>>',
    'arrive_time.end_time':'<<arrive time end time>>',
    'state_name':'<<state name>>',
    'airport_name':'<<airport name>>',
    'depart_date.day_name':'<<depart date day name>>',
    'fromloc.state_name':'<<fromloc state name>>',
    'arrive_time.time_relative':'<<arrive time time relative>>',
    'today_relative':'<<today relative>>',
    'day_name':'<<day name>>',
    'flight_stop':'<<flight stop>>',
    'month_name':'<<month name>>',
    'fromloc.city_name':'<<fromloc city name>>',
    'meal':'<<meal>>',
    'arrive_time.period_of_day':'<<arrive time period of day>>',
    'return_time.period_mod':'<<return time period mod>>',
    'toloc.airport_code':'<<toloc airport code>>',
    'airport_code':'<<airport code>>',
    'restriction_code':'<<restriction code>>',
    'flight_time':'<<flight time>>',
    'airline_code':'<<airline code>>',
    'depart_time.end_time':'<<depart time end time>>',
    'flight_days':'<<flight days>>',
    'booking_class':'<<booking class>>',
    'flight_number':'<<flight number>>',
    'or':'<<or>>',
    'fromloc.airport_code':'<<fromloc airport code>>',
    'meal_description':'<<meal description>>',
    'return_date.date_relative':'<<return date date relative>>',
    'return_date.month_name':'<<return date month name>>',
    'arrive_date.date_relative':'<<arrive date date relative>>',
    'return_date.today_relative':'<<return date today relative>>',
    'arrive_time.period_mod':'<<arrive time period mod>>',
    'depart_time.period_mod':'<<depart time period mod>>',
    'meal_code':'<<meal code>>',
    'flight':'<<flight>>',
    'toloc.airport_name':'<<toloc airport name>>',
    'stoploc.city_name':'<<stoploc city name>>',
    'connect':'<<connect>>',
    'arrive_time.time':'<<arrive time time>>',
    'toloc.state_code':'<<toloc state code>>',
    'fromloc.state_code':'<<fromloc state code>>'
}

mit_movie_mapping = {
    'genre': '<<genre>>',
    'actor': '<<actor>>',
    'year': '<<year>>',
    'title': '<<title>>',
    'rating': '<<rating>>',
    'ratings_average': '<<ratings average>>',
    'director': '<<director>>',
    'plot': '<<plot>>',
    'character': '<<character>>',
    'song': '<<song>>',
    'review': '<<review>>',
    'trailer': '<<trailer>>'
}

cluener2020_mapping = {
    'scene': '<<scene>>',
    'address': '<<address>>',
    'book': '<<book>>',
    'name': '<<name>>',
    'company': '<<company>>',
    'organization': '<<organization>>',
    'government': '<<government>>',
    'movie': '<<movie>>',
    'position': '<<position>>',
    'game': '<<game>>'
}

mit_restaurant_mapping = {
    'location': '<<location>>',
    'cuisine': '<<cuisine>>',
    'amenity': '<<amenity>>',
    'restaurant_name': '<<restaurant name>>',
    'rating': '<<rating>>',
    'dish': '<<dish>>',
    'hours': '<<hours>>',
    'price': '<<price>>'
}
